@@ -30,8 +30,8 @@
 
 #include "config.h"
 
-#include "galias.h"
 #include "glib.h"
+#include "galias.h"
 
 
 #define HASH_TABLE_MIN_SIZE 11
@@ -545,7 +545,7 @@ g_hash_table_foreach_remove_or_steal (GHashTable *hash_table,
  * pair, and the given @user_data parameter.  The hash table may not
  * be modified while iterating over it (you can't add/remove
  * items). To remove all items matching a predicate, use
- * g_hash_table_remove().
+ * g_hash_table_foreach_remove().
  **/
 void
 g_hash_table_foreach (GHashTable *hash_table,
@@ -759,3 +759,6 @@ g_hash_nodes_destroy (GHashNode *hash_node,
     }
 #endif
 }
+
+#define __G_HASH_C__
+#include "galiasdef.c"
