@@ -27,6 +27,8 @@
 
 #if defined(G_OS_UNIX) && !defined(G_STDIO_NO_WRAP_ON_UNIX)
 
+G_BEGIN_DECLS
+
 /* Just pass on to the system functions, so there's no potential for data
  * format mismatches, especially with large file interfaces.
  */
@@ -98,6 +100,8 @@ FILE *g_freopen (const gchar *filename,
                  const gchar *mode,
                  FILE        *stream);
 
+G_END_DECLS
+
 #endif /* G_OS_UNIX */
 
 #endif /* __G_STDIO_H__ */
