@@ -1,5 +1,6 @@
 /* GObject - GLib Type, Object, Parameter and Signal Library
  * Copyright (C) 2000-2001 Red Hat, Inc.
+ * Copyright (C) 2005 Imendio AB
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -18,14 +19,15 @@
  */
 #include	"gclosure.h"
 
+/*
+ * MT safe with regards to reference counting.
+ */
+
 #include	"gvalue.h"
 #include 	"gobjectalias.h"
 #include	<string.h>
 
 
-/* FIXME: need caching allocators
- */
-
 #define	CLOSURE_MAX_REF_COUNT		((1 << 15) - 1)
 #define	CLOSURE_MAX_N_GUARDS		((1 << 1) - 1)
 #define	CLOSURE_MAX_N_FNOTIFIERS	((1 << 2) - 1)
@@ -37,42 +39,43 @@
                                          (cl)->n_fnotifiers + \
                                          (cl)->n_inotifiers)
 
-/* union of first int we need to make atomic */
 typedef union {
-  GClosure bits;
-  gint atomic;
-} GAtomicClosureBits;
+  GClosure closure;
+  volatile gint vint;
+} ClosureInt;
 
-#define BITS_AS_INT(b)	(((GAtomicClosureBits*)(b))->atomic)
-
-#define CLOSURE_READ_BITS(cl,bits)	(BITS_AS_INT(bits) = g_atomic_int_get ((gint*)(cl)))
-#define CLOSURE_READ_BITS2(cl,old,new)	(BITS_AS_INT(old) = CLOSURE_READ_BITS (cl, new))
-#define CLOSURE_SWAP_BITS(cl,old,new)	(g_atomic_int_compare_and_exchange ((gint*)(cl),	\
-						BITS_AS_INT(old),BITS_AS_INT(new)))
-
-#define CLOSURE_REF(closure)  							\
-G_STMT_START {									\
-  GClosure old, new;								\
-  do {										\
-    CLOSURE_READ_BITS2 (closure, &old, &new);					\
-    new.ref_count++;								\
-  }										\
-  while (!CLOSURE_SWAP_BITS (closure, &old, &new));				\
+#define CHANGE_FIELD(_closure, _field, _OP, _value, _must_set, _SET_OLD, _SET_NEW)      \
+G_STMT_START {                                                                          \
+  ClosureInt *cunion = (ClosureInt*) _closure;                 		                \
+  gint new_int, old_int, success;                              		                \
+  do                                                    		                \
+    {                                                   		                \
+      ClosureInt tmp;                                   		                \
+      tmp.vint = old_int = cunion->vint;                		                \
+      _SET_OLD tmp.closure._field;                                                      \
+      tmp.closure._field _OP _value;                      		                \
+      _SET_NEW tmp.closure._field;                                                      \
+      new_int = tmp.vint;                               		                \
+      success = g_atomic_int_compare_and_exchange (&cunion->vint, old_int, new_int);    \
+    }                                                   		                \
+  while (!success && _must_set);                                                        \
 } G_STMT_END
 
+#define SWAP(_closure, _field, _value, _oldv)   CHANGE_FIELD (_closure, _field, =, _value, TRUE, *(_oldv) =,     (void) )
+#define SET(_closure, _field, _value)           CHANGE_FIELD (_closure, _field, =, _value, TRUE,     (void),     (void) )
+#define INC(_closure, _field)                   CHANGE_FIELD (_closure, _field, +=,     1, TRUE,     (void),     (void) )
+#define INC_ASSIGN(_closure, _field, _newv)     CHANGE_FIELD (_closure, _field, +=,     1, TRUE,     (void), *(_newv) = )
+#define DEC(_closure, _field)                   CHANGE_FIELD (_closure, _field, -=,     1, TRUE,     (void),     (void) )
+#define DEC_ASSIGN(_closure, _field, _newv)     CHANGE_FIELD (_closure, _field, -=,     1, TRUE,     (void), *(_newv) = )
 
-#define CLOSURE_UNREF(closure, is_zero)  					\
-G_STMT_START {									\
-  GClosure old, new;								\
-  do {										\
-    CLOSURE_READ_BITS2 (closure, &old, &new);					\
-    if (old.ref_count == 1)	/* last unref, invalidate first */		\
-      g_closure_invalidate ((closure));						\
-    new.ref_count--;								\
-    is_zero = (new.ref_count == 0);						\
-  }										\
-  while (!CLOSURE_SWAP_BITS (closure, &old, &new));				\
-} G_STMT_END
+#if 0   // for non-thread-safe closures
+#define SWAP(cl,f,v,o)     (void) (*(o) = cl->f, cl->f = v)
+#define SET(cl,f,v)        (void) (cl->f = v)
+#define INC(cl,f)          (void) (cl->f += 1)
+#define INC_ASSIGN(cl,f,n) (void) (cl->f += 1, *(n) = cl->f)
+#define DEC(cl,f)          (void) (cl->f -= 1)
+#define DEC_ASSIGN(cl,f,n) (void) (cl->f -= 1, *(n) = cl->f)
+#endif
 
 enum {
   FNOTIFY,
@@ -91,17 +94,17 @@ g_closure_new_simple (guint           sizeof_closure,
 
   g_return_val_if_fail (sizeof_closure >= sizeof (GClosure), NULL);
 
-  closure = g_malloc (sizeof_closure);
-  closure->ref_count = 1;
-  closure->meta_marshal = 0;
-  closure->n_guards = 0;
-  closure->n_fnotifiers = 0;
-  closure->n_inotifiers = 0;
-  closure->in_inotify = FALSE;
-  closure->floating = TRUE;
-  closure->derivative_flag = 0;
-  closure->in_marshal = FALSE;
-  closure->is_invalid = FALSE;
+  closure = g_malloc0 (sizeof_closure);
+  SET (closure, ref_count, 1);
+  SET (closure, meta_marshal, 0);
+  SET (closure, n_guards, 0);
+  SET (closure, n_fnotifiers, 0);
+  SET (closure, n_inotifiers, 0);
+  SET (closure, in_inotify, FALSE);
+  SET (closure, floating, TRUE);
+  SET (closure, derivative_flag, 0);
+  SET (closure, in_marshal, FALSE);
+  SET (closure, is_invalid, FALSE);
   closure->marshal = NULL;
   closure->data = data;
   closure->notifiers = NULL;
@@ -114,8 +117,6 @@ static inline void
 closure_invoke_notifiers (GClosure *closure,
 			  guint     notify_type)
 {
-  GClosure bits, new;
-
   /* notifier layout:
    *     meta_marshal  n_guards    n_guards     n_fnotif.  n_inotifiers
    * ->[[meta_marshal][pre_guards][post_guards][fnotifiers][inotifiers]]
@@ -127,8 +128,8 @@ closure_invoke_notifiers (GClosure *closure,
    * - closure->notifiers may be reloacted during callback
    * - closure->n_fnotifiers and closure->n_inotifiers may change during callback
    * - i.e. callbacks can be removed/added during invocation
-   * - have to prepare for callback removal during invocation (->marshal & ->data)
-   * - have to distinguish (->marshal & ->data) for INOTIFY/FNOTIFY (->in_inotify)
+   * - must prepare for callback removal during FNOTIFY and INOTIFY (done via ->marshal= & ->data=)
+   * - must distinguish (->marshal= & ->data=) for INOTIFY vs. FNOTIFY (via ->in_inotify)
    * + closure->n_guards is const during PRE_NOTIFY & POST_NOTIFY
    * + closure->meta_marshal is const for all cases
    * + none of the callbacks can cause recursion
@@ -139,12 +140,12 @@ closure_invoke_notifiers (GClosure *closure,
       GClosureNotifyData *ndata;
       guint i, offs;
     case FNOTIFY:
-      CLOSURE_READ_BITS (closure, &bits);
-      while (bits.n_fnotifiers)
+      while (closure->n_fnotifiers)
 	{
-	  register guint n = --bits.n_fnotifiers;
+          guint n;
+	  DEC_ASSIGN (closure, n_fnotifiers, &n);
 
-	  ndata = closure->notifiers + CLOSURE_N_MFUNCS (&bits) + n;
+	  ndata = closure->notifiers + CLOSURE_N_MFUNCS (closure) + n;
 	  closure->marshal = (GClosureMarshal) ndata->notify;
 	  closure->data = ndata->data;
 	  ndata->notify (ndata->data, closure);
@@ -153,34 +154,24 @@ closure_invoke_notifiers (GClosure *closure,
       closure->data = NULL;
       break;
     case INOTIFY:
-      do {
-        CLOSURE_READ_BITS2 (closure, &bits, &new);
-        new.in_inotify = TRUE;
-      }
-      while (!CLOSURE_SWAP_BITS (closure, &bits,  &new));
-
-      while (bits.n_inotifiers)
+      SET (closure, in_inotify, TRUE);
+      while (closure->n_inotifiers)
 	{
-          register guint n = --bits.n_inotifiers;
+          guint n;
+          DEC_ASSIGN (closure, n_inotifiers, &n);
 
-	  ndata = closure->notifiers + CLOSURE_N_MFUNCS (&bits) + bits.n_fnotifiers + n;
+	  ndata = closure->notifiers + CLOSURE_N_MFUNCS (closure) + closure->n_fnotifiers + n;
 	  closure->marshal = (GClosureMarshal) ndata->notify;
 	  closure->data = ndata->data;
 	  ndata->notify (ndata->data, closure);
 	}
       closure->marshal = NULL;
       closure->data = NULL;
-      do {
-        CLOSURE_READ_BITS2 (closure, &bits, &new);
-        new.n_inotifiers = 0;
-        new.in_inotify = FALSE;
-      }
-      while (!CLOSURE_SWAP_BITS (closure, &bits, &new));
+      SET (closure, in_inotify, FALSE);
       break;
     case PRE_NOTIFY:
-      CLOSURE_READ_BITS (closure, &bits);
-      i = bits.n_guards;
-      offs = bits.meta_marshal;
+      i = closure->n_guards;
+      offs = closure->meta_marshal;
       while (i--)
 	{
 	  ndata = closure->notifiers + offs + i;
@@ -188,9 +179,8 @@ closure_invoke_notifiers (GClosure *closure,
 	}
       break;
     case POST_NOTIFY:
-      CLOSURE_READ_BITS (closure, &bits);
-      i = bits.n_guards;
-      offs = bits.meta_marshal + i;
+      i = closure->n_guards;
+      offs = closure->meta_marshal + i;
       while (i--)
 	{
 	  ndata = closure->notifiers + offs + i;
@@ -205,48 +195,29 @@ g_closure_set_meta_marshal (GClosure       *closure,
 			    gpointer        marshal_data,
 			    GClosureMarshal meta_marshal)
 {
-  GClosureNotifyData *old_notifiers, *new_notifiers;
+  GClosureNotifyData *notifiers;
   guint n;
-  GClosure old, new;
 
   g_return_if_fail (closure != NULL);
   g_return_if_fail (meta_marshal != NULL);
+  g_return_if_fail (closure->is_invalid == FALSE);
+  g_return_if_fail (closure->in_marshal == FALSE);
+  g_return_if_fail (closure->meta_marshal == 0);
 
-retry:
-  CLOSURE_READ_BITS2 (closure, &old, &new);
-
-  g_return_if_fail (old.is_invalid == FALSE);
-  g_return_if_fail (old.in_marshal == FALSE);
-  g_return_if_fail (old.meta_marshal == 0);
-
-  n = CLOSURE_N_NOTIFIERS (&old);
-
-  old_notifiers = closure->notifiers;
-  new_notifiers = g_renew (GClosureNotifyData, NULL, n + 1);
-  if (old_notifiers)
+  n = CLOSURE_N_NOTIFIERS (closure);
+  notifiers = closure->notifiers;
+  closure->notifiers = g_renew (GClosureNotifyData, NULL, CLOSURE_N_NOTIFIERS (closure) + 1);
+  if (notifiers)
     {
       /* usually the meta marshal will be setup right after creation, so the
        * g_memmove() should be rare-case scenario
        */
-      g_memmove (new_notifiers + 1, old_notifiers, n * sizeof (old_notifiers[0]));
+      g_memmove (closure->notifiers + 1, notifiers, CLOSURE_N_NOTIFIERS (closure) * sizeof (notifiers[0]));
+      g_free (notifiers);
     }
-  new_notifiers[0].data = marshal_data;
-  new_notifiers[0].notify = (GClosureNotify) meta_marshal;
-
-  new.meta_marshal = 1;
-
-  /* this cannot be made atomic, as soon as we switch on the meta_marshal
-   * bit, another thread could use the notifier while we have not yet
-   * copied it. the safest is to install the new_notifiers first and then
-   * switch on the meta_marshal flag. */
-  closure->notifiers = new_notifiers;
-
-  if (!CLOSURE_SWAP_BITS (closure, &old, &new)) {
-    g_free (new_notifiers);
-    goto retry;
-  }
-
-  g_free (old_notifiers);
+  closure->notifiers[0].data = marshal_data;
+  closure->notifiers[0].notify = (GClosureNotify) meta_marshal;
+  SET (closure, meta_marshal, 1);
 }
 
 void
@@ -257,56 +228,41 @@ g_closure_add_marshal_guards (GClosure      *closure,
 			      GClosureNotify post_marshal_notify)
 {
   guint i;
-  GClosure old, new;
-  GClosureNotifyData *old_notifiers, *new_notifiers;
 
   g_return_if_fail (closure != NULL);
   g_return_if_fail (pre_marshal_notify != NULL);
   g_return_if_fail (post_marshal_notify != NULL);
+  g_return_if_fail (closure->is_invalid == FALSE);
+  g_return_if_fail (closure->in_marshal == FALSE);
+  g_return_if_fail (closure->n_guards < CLOSURE_MAX_N_GUARDS);
 
-retry:
-  CLOSURE_READ_BITS2 (closure,  &old,  &new);
-
-  g_return_if_fail (old.is_invalid == FALSE);
-  g_return_if_fail (old.in_marshal == FALSE);
-  g_return_if_fail (old.n_guards < CLOSURE_MAX_N_GUARDS);
-
-  old_notifiers = closure->notifiers;
-  new_notifiers = g_renew (GClosureNotifyData, old_notifiers, CLOSURE_N_NOTIFIERS (&old) + 2);
-  if (old.n_inotifiers)
-    new_notifiers[(CLOSURE_N_MFUNCS (&old) +
-			old.n_fnotifiers +
-			old.n_inotifiers + 1)] = new_notifiers[(CLOSURE_N_MFUNCS (&old) +
-									  old.n_fnotifiers + 0)];
-  if (old.n_inotifiers > 1)
-    new_notifiers[(CLOSURE_N_MFUNCS (&old) +
-			old.n_fnotifiers +
-			old.n_inotifiers)] = new_notifiers[(CLOSURE_N_MFUNCS (&old) +
-								      old.n_fnotifiers + 1)];
-  if (old.n_fnotifiers)
-    new_notifiers[(CLOSURE_N_MFUNCS (&old) +
-			old.n_fnotifiers + 1)] = new_notifiers[CLOSURE_N_MFUNCS (&old) + 0];
-  if (old.n_fnotifiers > 1)
-    new_notifiers[(CLOSURE_N_MFUNCS (&old) +
-			old.n_fnotifiers)] = new_notifiers[CLOSURE_N_MFUNCS (&old) + 1];
-  if (old.n_guards)
-    new_notifiers[(old.meta_marshal +
-			old.n_guards +
-			old.n_guards + 1)] = new_notifiers[old.meta_marshal + old.n_guards];
-  i = old.n_guards;
-
-  new.n_guards = i+1;
-
-  new_notifiers[old.meta_marshal + i].data = pre_marshal_data;
-  new_notifiers[old.meta_marshal + i].notify = pre_marshal_notify;
-  new_notifiers[old.meta_marshal + i + 1].data = post_marshal_data;
-  new_notifiers[old.meta_marshal + i + 1].notify = post_marshal_notify;
-
-  /* not really atomic */
-  closure->notifiers = new_notifiers;
-
-  if (!CLOSURE_SWAP_BITS (closure, &old, &new))
-    goto retry;
+  closure->notifiers = g_renew (GClosureNotifyData, closure->notifiers, CLOSURE_N_NOTIFIERS (closure) + 2);
+  if (closure->n_inotifiers)
+    closure->notifiers[(CLOSURE_N_MFUNCS (closure) +
+			closure->n_fnotifiers +
+			closure->n_inotifiers + 1)] = closure->notifiers[(CLOSURE_N_MFUNCS (closure) +
+									  closure->n_fnotifiers + 0)];
+  if (closure->n_inotifiers > 1)
+    closure->notifiers[(CLOSURE_N_MFUNCS (closure) +
+			closure->n_fnotifiers +
+			closure->n_inotifiers)] = closure->notifiers[(CLOSURE_N_MFUNCS (closure) +
+								      closure->n_fnotifiers + 1)];
+  if (closure->n_fnotifiers)
+    closure->notifiers[(CLOSURE_N_MFUNCS (closure) +
+			closure->n_fnotifiers + 1)] = closure->notifiers[CLOSURE_N_MFUNCS (closure) + 0];
+  if (closure->n_fnotifiers > 1)
+    closure->notifiers[(CLOSURE_N_MFUNCS (closure) +
+			closure->n_fnotifiers)] = closure->notifiers[CLOSURE_N_MFUNCS (closure) + 1];
+  if (closure->n_guards)
+    closure->notifiers[(closure->meta_marshal +
+			closure->n_guards +
+			closure->n_guards + 1)] = closure->notifiers[closure->meta_marshal + closure->n_guards];
+  i = closure->n_guards;
+  closure->notifiers[closure->meta_marshal + i].data = pre_marshal_data;
+  closure->notifiers[closure->meta_marshal + i].notify = pre_marshal_notify;
+  closure->notifiers[closure->meta_marshal + i + 1].data = post_marshal_data;
+  closure->notifiers[closure->meta_marshal + i + 1].notify = post_marshal_notify;
+  INC (closure, n_guards);
 }
 
 void
@@ -315,35 +271,21 @@ g_closure_add_finalize_notifier (GClosure      *closure,
 				 GClosureNotify notify_func)
 {
   guint i;
-  GClosure old, new;
-  GClosureNotifyData *old_notifiers, *new_notifiers;
 
   g_return_if_fail (closure != NULL);
   g_return_if_fail (notify_func != NULL);
+  g_return_if_fail (closure->n_fnotifiers < CLOSURE_MAX_N_FNOTIFIERS);
 
-retry:
-  CLOSURE_READ_BITS2 (closure, &old, &new);
-
-  g_return_if_fail (old.n_fnotifiers < CLOSURE_MAX_N_FNOTIFIERS);
-
-  old_notifiers = closure->notifiers;
-  new_notifiers = g_renew (GClosureNotifyData, old_notifiers, CLOSURE_N_NOTIFIERS (&old) + 1);
-  if (old.n_inotifiers)
-    new_notifiers[(CLOSURE_N_MFUNCS (&old) +
-			old.n_fnotifiers +
-			old.n_inotifiers)] = new_notifiers[(CLOSURE_N_MFUNCS (&old) +
-								      old.n_fnotifiers + 0)];
-  i = CLOSURE_N_MFUNCS (&old) + old.n_fnotifiers;
-  new.n_fnotifiers++;
-
-  new_notifiers[i].data = notify_data;
-  new_notifiers[i].notify = notify_func;
-
-  /* not really atomic */
-  closure->notifiers = new_notifiers;
-
-  while (!CLOSURE_SWAP_BITS (closure, &old, &new))
-    goto retry;
+  closure->notifiers = g_renew (GClosureNotifyData, closure->notifiers, CLOSURE_N_NOTIFIERS (closure) + 1);
+  if (closure->n_inotifiers)
+    closure->notifiers[(CLOSURE_N_MFUNCS (closure) +
+			closure->n_fnotifiers +
+			closure->n_inotifiers)] = closure->notifiers[(CLOSURE_N_MFUNCS (closure) +
+								      closure->n_fnotifiers + 0)];
+  i = CLOSURE_N_MFUNCS (closure) + closure->n_fnotifiers;
+  closure->notifiers[i].data = notify_data;
+  closure->notifiers[i].notify = notify_func;
+  INC (closure, n_fnotifiers);
 }
 
 void
@@ -352,31 +294,17 @@ g_closure_add_invalidate_notifier (GClosure      *closure,
 				   GClosureNotify notify_func)
 {
   guint i;
-  GClosure old, new;
-  GClosureNotifyData *old_notifiers, *new_notifiers;
 
   g_return_if_fail (closure != NULL);
   g_return_if_fail (notify_func != NULL);
+  g_return_if_fail (closure->is_invalid == FALSE);
+  g_return_if_fail (closure->n_inotifiers < CLOSURE_MAX_N_INOTIFIERS);
 
-retry:
-  CLOSURE_READ_BITS2 (closure, &old, &new);
-
-  g_return_if_fail (old.is_invalid == FALSE);
-  g_return_if_fail (old.n_inotifiers < CLOSURE_MAX_N_INOTIFIERS);
-
-  old_notifiers = closure->notifiers;
-  new_notifiers = g_renew (GClosureNotifyData, old_notifiers, CLOSURE_N_NOTIFIERS (&old) + 1);
-  i = CLOSURE_N_MFUNCS (&old) + old.n_fnotifiers + old.n_inotifiers;
-  new.n_inotifiers++;
-
-  new_notifiers[i].data = notify_data;
-  new_notifiers[i].notify = notify_func;
-
-  /* not really atomic */
-  closure->notifiers = new_notifiers;
-
-  while (!CLOSURE_SWAP_BITS (closure, &old, &new))
-    goto retry;
+  closure->notifiers = g_renew (GClosureNotifyData, closure->notifiers, CLOSURE_N_NOTIFIERS (closure) + 1);
+  i = CLOSURE_N_MFUNCS (closure) + closure->n_fnotifiers + closure->n_inotifiers;
+  closure->notifiers[i].data = notify_data;
+  closure->notifiers[i].notify = notify_func;
+  INC (closure, n_inotifiers);
 }
 
 static inline gboolean
@@ -385,19 +313,12 @@ closure_try_remove_inotify (GClosure       *closure,
 			    GClosureNotify notify_func)
 {
   GClosureNotifyData *ndata, *nlast;
-  GClosure old, new;
 
-retry:
-  CLOSURE_READ_BITS2 (closure, &old, &new);
-
-  nlast = closure->notifiers + CLOSURE_N_NOTIFIERS (&old) - 1;
-  for (ndata = nlast + 1 - old.n_inotifiers; ndata <= nlast; ndata++)
+  nlast = closure->notifiers + CLOSURE_N_NOTIFIERS (closure) - 1;
+  for (ndata = nlast + 1 - closure->n_inotifiers; ndata <= nlast; ndata++)
     if (ndata->notify == notify_func && ndata->data == notify_data)
       {
-	new.n_inotifiers -= 1;
-	if (!CLOSURE_SWAP_BITS (closure, &old, &new))
-          goto retry;
-	
+	DEC (closure, n_inotifiers);
 	if (ndata < nlast)
 	  *ndata = *nlast;
 
@@ -412,27 +333,19 @@ closure_try_remove_fnotify (GClosure       *closure,
 			    GClosureNotify notify_func)
 {
   GClosureNotifyData *ndata, *nlast;
-  GClosure old, new;
 
-retry:
-  CLOSURE_READ_BITS2 (closure, &old, &new);
-
-  nlast = closure->notifiers + CLOSURE_N_NOTIFIERS (&old) - old.n_inotifiers - 1;
-  for (ndata = nlast + 1 - old.n_fnotifiers; ndata <= nlast; ndata++)
+  nlast = closure->notifiers + CLOSURE_N_NOTIFIERS (closure) - closure->n_inotifiers - 1;
+  for (ndata = nlast + 1 - closure->n_fnotifiers; ndata <= nlast; ndata++)
     if (ndata->notify == notify_func && ndata->data == notify_data)
       {
-	new.n_fnotifiers -= 1;
-	if (!CLOSURE_SWAP_BITS (closure, &old, &new))
-          goto retry;
-	
+	DEC (closure, n_fnotifiers);
 	if (ndata < nlast)
 	  *ndata = *nlast;
-
-	if (new.n_inotifiers)
-	  closure->notifiers[(CLOSURE_N_MFUNCS (&new) +
-			      new.n_fnotifiers)] = closure->notifiers[(CLOSURE_N_MFUNCS (&new) +
-									    new.n_fnotifiers +
-									    new.n_inotifiers)];
+	if (closure->n_inotifiers)
+	  closure->notifiers[(CLOSURE_N_MFUNCS (closure) +
+			      closure->n_fnotifiers)] = closure->notifiers[(CLOSURE_N_MFUNCS (closure) +
+									    closure->n_fnotifiers +
+									    closure->n_inotifiers)];
 	return TRUE;
       }
   return FALSE;
@@ -441,11 +354,13 @@ retry:
 GClosure*
 g_closure_ref (GClosure *closure)
 {
+  guint new_ref_count;
   g_return_val_if_fail (closure != NULL, NULL);
   g_return_val_if_fail (closure->ref_count > 0, NULL);
   g_return_val_if_fail (closure->ref_count < CLOSURE_MAX_REF_COUNT, NULL);
 
-  CLOSURE_REF (closure);
+  INC_ASSIGN (closure, ref_count, &new_ref_count);
+  g_return_val_if_fail (new_ref_count > 1, NULL);
 
   return closure;
 }
@@ -453,22 +368,16 @@ g_closure_ref (GClosure *closure)
 void
 g_closure_invalidate (GClosure *closure)
 {
-  GClosure old, new;
-
   g_return_if_fail (closure != NULL);
 
-retry:
-  CLOSURE_READ_BITS2 (closure, &old, &new);
-
-  if (!old.is_invalid)
+  if (!closure->is_invalid)
     {
-      new.ref_count++;
-      new.is_invalid = TRUE;
-
-      if (!CLOSURE_SWAP_BITS (closure, &old, &new))
-	goto retry;
-
-      closure_invoke_notifiers (closure, INOTIFY);
+      gboolean was_invalid;
+      g_closure_ref (closure);           /* preserve floating flag */
+      SWAP (closure, is_invalid, TRUE, &was_invalid);
+      /* invalidate only once */
+      if (!was_invalid)
+        closure_invoke_notifiers (closure, INOTIFY);
       g_closure_unref (closure);
     }
 }
@@ -476,14 +385,17 @@ retry:
 void
 g_closure_unref (GClosure *closure)
 {
-  gboolean is_zero;
+  guint new_ref_count;
 
   g_return_if_fail (closure != NULL);
   g_return_if_fail (closure->ref_count > 0);
 
-  CLOSURE_UNREF (closure, is_zero);
+  if (closure->ref_count == 1)	/* last unref, invalidate first */
+    g_closure_invalidate (closure);
 
-  if (G_UNLIKELY (is_zero))
+  DEC_ASSIGN (closure, ref_count, &new_ref_count);
+
+  if (new_ref_count == 0)
     {
       closure_invoke_notifiers (closure, FNOTIFY);
       g_free (closure->notifiers);
@@ -494,8 +406,6 @@ g_closure_unref (GClosure *closure)
 void
 g_closure_sink (GClosure *closure)
 {
-  GClosure old, new;
-
   g_return_if_fail (closure != NULL);
   g_return_if_fail (closure->ref_count > 0);
 
@@ -504,17 +414,13 @@ g_closure_sink (GClosure *closure)
    * is unowned. with invoking g_closure_sink() code may
    * indicate that it takes over that intiial ref_count.
    */
-retry:
-  CLOSURE_READ_BITS2 (closure, &old, &new);
-
-  if (old.floating)
+  if (closure->floating)
     {
-      new.floating = FALSE;
-
-      if (!CLOSURE_SWAP_BITS (closure, &old, &new))
-	goto retry;
-
-      g_closure_unref (closure);
+      gboolean was_floating;
+      SWAP (closure, floating, FALSE, &was_floating);
+      /* unref floating flag only once */
+      if (was_floating)
+        g_closure_unref (closure);
     }
 }
 
@@ -523,15 +429,12 @@ g_closure_remove_invalidate_notifier (GClosure      *closure,
 				      gpointer       notify_data,
 				      GClosureNotify notify_func)
 {
-  GClosure bits;
-
   g_return_if_fail (closure != NULL);
   g_return_if_fail (notify_func != NULL);
 
-  CLOSURE_READ_BITS (closure, &bits);
-
-  if (bits.is_invalid && bits.in_inotify && /* account removal of notify_func() while its called */
-      ((gpointer) closure->marshal) == ((gpointer) notify_func) && closure->data == notify_data)
+  if (closure->is_invalid && closure->in_inotify && /* account removal of notify_func() while its called */
+      ((gpointer) closure->marshal) == ((gpointer) notify_func) &&
+      closure->data == notify_data)
     closure->marshal = NULL;
   else if (!closure_try_remove_inotify (closure, notify_data, notify_func))
     g_warning (G_STRLOC ": unable to remove uninstalled invalidation notifier: %p (%p)",
@@ -543,15 +446,12 @@ g_closure_remove_finalize_notifier (GClosure      *closure,
 				    gpointer       notify_data,
 				    GClosureNotify notify_func)
 {
-  GClosure bits;
-	
   g_return_if_fail (closure != NULL);
   g_return_if_fail (notify_func != NULL);
 
-  CLOSURE_READ_BITS (closure, &bits);
-
-  if (bits.is_invalid && !bits.in_inotify && /* account removal of notify_func() while its called */
-      ((gpointer) closure->marshal) == ((gpointer) notify_func) && closure->data == notify_data)
+  if (closure->is_invalid && !closure->in_inotify && /* account removal of notify_func() while its called */
+      ((gpointer) closure->marshal) == ((gpointer) notify_func) &&
+      closure->data == notify_data)
     closure->marshal = NULL;
   else if (!closure_try_remove_fnotify (closure, notify_data, notify_func))
     g_warning (G_STRLOC ": unable to remove uninstalled finalization notifier: %p (%p)",
@@ -565,29 +465,19 @@ g_closure_invoke (GClosure       *closure,
 		  const GValue   *param_values,
 		  gpointer        invocation_hint)
 {
-  GClosure old, new;
-
   g_return_if_fail (closure != NULL);
 
-retry:
-  CLOSURE_READ_BITS2 (closure, &old, &new);
-
-  if (!old.is_invalid)
-   {
+  g_closure_ref (closure);      /* preserve floating flag */
+  if (!closure->is_invalid)
+    {
       GClosureMarshal marshal;
       gpointer marshal_data;
-      gboolean in_marshal = old.in_marshal;
-      gboolean meta_marshal = old.meta_marshal;
+      gboolean in_marshal = closure->in_marshal;
 
-      g_return_if_fail (closure->marshal || meta_marshal);
+      g_return_if_fail (closure->marshal || closure->meta_marshal);
 
-      new.ref_count++;
-      new.in_marshal = TRUE;
-
-      if (!CLOSURE_SWAP_BITS (closure, &old, &new))
-	goto retry;
-
-      if (meta_marshal)
+      SET (closure, in_marshal, TRUE);
+      if (closure->meta_marshal)
 	{
 	  marshal_data = closure->notifiers[0].data;
 	  marshal = (GClosureMarshal) closure->notifiers[0].notify;
@@ -599,23 +489,16 @@ retry:
 	}
       if (!in_marshal)
 	closure_invoke_notifiers (closure, PRE_NOTIFY);
-
       marshal (closure,
 	       return_value,
 	       n_param_values, param_values,
 	       invocation_hint,
 	       marshal_data);
-
       if (!in_marshal)
 	closure_invoke_notifiers (closure, POST_NOTIFY);
-
-      do {
-        CLOSURE_READ_BITS2 (closure, &old, &new);
-        new.in_marshal = in_marshal;
-        new.ref_count--;
-      }
-      while (!CLOSURE_SWAP_BITS (closure, &old, &new));
+      SET (closure, in_marshal, in_marshal);
     }
+  g_closure_unref (closure);
 }
 
 void
@@ -662,7 +545,7 @@ g_cclosure_new_swap (GCallback      callback_func,
   if (destroy_data)
     g_closure_add_finalize_notifier (closure, user_data, destroy_data);
   ((GCClosure*) closure)->callback = (gpointer) callback_func;
-  closure->derivative_flag = TRUE;
+  SET (closure, derivative_flag, TRUE);
   
   return closure;
 }
